"""
a plugin to generate a sitemap

http://en.wikipedia.org/wiki/Sitemaps
"""
from __future__ import absolute_import
import os
import codecs

from .. import event, echo

def output_sitemap(event_name, site):
    if not len(site.posts): return

    sitemap = os.path.join(str(site.output_dir), 'sitemap.xml')
    echo.out("writing sitemap to {}", sitemap)

    host = site.config.host
    max_count = 50000
    count = 0

    if host:
        with codecs.open(sitemap, 'w+', 'utf-8') as fp:
            fp.write(u"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
            fp.write(u"<urlset xmlns=\"http://www.sitemaps.org/schemas/sitemap/0.9\">\n")

            for p in reversed(site.posts):
                fp.write(u"  <url>\n")
                fp.write(u"    <loc>{}</loc>\n".format(p.url))
                fp.write(u"    <lastmod>{}</lastmod>\n".format(p.modified.strftime("%Y-%m-%dT%H:%M:%S+00:00")))
                fp.write(u"    <changefreq>weekly</changefreq>\n")
                #fp.write(u'    <priority>0.8</priority>')
                fp.write(u"  </url>\n")

                count += 1
                if count >= max_count:
                    break

            fp.write(u"</urlset>\n")

    else:
        echo.err("[WARNING] sitemap not generated because no config host set")


event.listen('output.finish', output_sitemap)




from . import echo

events = {}

def listen(event_name, callback):
    global events

    events.setdefault(event_name, [])
    events[event_name].append(callback)


def broadcast(event_name, *args, **kwargs):
    global events

    callbacks = events.get(event_name, [])
    echo.out("broadcast event {} to {} callbacks", event_name, len(callbacks))

    for callback in callbacks:
        callback(event_name, *args, **kwargs)




"""
Why do I render the feed myself? Because every library I found needed to many other
dependencies and some were a real pain to install (you need lxml? Seriously?). So
I just render the feed raw

http://en.wikipedia.org/wiki/RSS
https://github.com/lkiesow/python-feedgen
http://cyber.law.harvard.edu/rss/rss.html

alidator: http://validator.w3.org/feed/
big list of namespaces: http://validator.w3.org/feed/docs/howto/declare_namespaces.html
"""
import datetime
import os
from xml.sax.saxutils import escape
import codecs

from .. import event, echo


def get_safe(text):
    """return xml safe content with things like & < > escaped"""
    # https://wiki.python.org/moin/EscapingXml
    return escape(text)


def get_cdata(text):
    """wrap the text in cdata escaping"""
    return u'<![CDATA[{}]]>'.format(text)


def get_datestr(dt):
    """return a date from datetime like Mon, 06 Sep 2010 00:01:00 +0000"""
    return dt.strftime("%a, %d %b %Y %H:%M:%S +0000")


def output_rss(event_name, site):
    host = site.config.host
    if not host:
        echo.err("[WARNING] rss feed not generated because no config host set")
        return

    feedpath = os.path.join(str(site.output_dir), 'feed.rss')
    echo.out("writing feed to {}", feedpath)

    main_url = site.config.base_url
    feed_url = u'http://{}/feed.rss'.format(host)
    max_count = 10
    count = 0

    with codecs.open(feedpath, 'w+', 'utf-8') as fp:

        fp.write(u"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
        fp.write(u"<rss version=\"2.0\"\n")
        fp.write(u"  xmlns:atom=\"http://www.w3.org/2005/Atom\">\n")
        #fp.write(u"  xmlns:dc=\"http://purl.org/dc/elements/1.1/\"\n") # http://en.wikipedia.org/wiki/Dublin_Core
        #fp.write(u"  xmlns:dcterms=\"http://purl.org/dc/terms/\"\n")
        #fp.write(u"  xmlns:georss=\"http://www.georss.org/georss\">\n")

        fp.write(u"  <channel>\n")
        fp.write(u"    <title>{}</title>\n".format(get_safe(site.config.get('name', host))))
        fp.write(u"    <description>{}</description>\n".format(get_safe(site.config.get('description', host))))

        fp.write(u"    <link>{}</link>\n".format(get_safe(main_url)))
        fp.write(u"    <atom:link href=\"{}\" rel=\"self\"/>\n".format(get_safe(feed_url)))
        #fp.write(u"    <atom:link href=\"{}\" rel=\"alternate\"/>\n".format(main_url))

        dt = datetime.datetime.utcnow()
        fp.write(u"    <pubDate>{}</pubDate>\n".format(get_datestr(dt)))
        fp.write(u"    <lastBuildDate>{}</lastBuildDate>\n".format(get_datestr(dt)))
        fp.write(u"    <generator>github.com/Jaymon/bang</generator>\n")

        for p in reversed(site.posts):
            fp.write(u"    <item>\n")
            fp.write(u"      <title>{}</title>\n".format(get_cdata(p.title)))
            fp.write(u"      <description>{}</description>\n".format(get_cdata(p.html)))
            fp.write(u"      <link>{}</link>\n".format(get_safe(p.url)))
            fp.write(u"      <guid isPermaLink=\"false\">{}</guid>\n".format(get_safe(p.uri)))
            fp.write(u"      <pubDate>{}</pubDate>\n".format(get_datestr(p.modified)))
            fp.write(u"    </item>\n")

            count += 1
            if count >= max_count:
                break

        fp.write(u"  </channel>\n")
        fp.write(u"</rss>\n")


event.listen('output.finish', output_rss)




"""
in this module, we fix all the problems with markdown's default code display stuff

extension for the markdown lib I use: https://github.com/waylan/Python-Markdown

http://pythonhosted.org/Markdown/extensions/api.html

https://github.com/waylan/Python-Markdown/wiki/Third-Party-Extensions
"""
import re
import os

from markdown.extensions import codehilite, fenced_code
from pygments import highlight
from pygments.lexers import get_lexer_by_name, guess_lexer, TextLexer
from pygments.formatters import HtmlFormatter

from markdown.extensions import Extension
from markdown.treeprocessors import Treeprocessor
from . import event

class DomEventTreeprocessor(Treeprocessor):
    """support for HrefExtension"""
    def iterparent(self, tree):
        for parent in tree.getiterator():
            for child in parent:
                yield parent, child

    def run(self, doc):
        post = self.config['post']
        for parent, elem in self.iterparent(doc):
            elem_event_name = 'dom.{}'.format(elem.tag)
            event.broadcast(elem_event_name, parent=parent, elem=elem, **self.config)


class DomEventExtension(Extension):
    """
    this will modify all a href attributes to be full url paths

    if it is an http://... url, nothing happens, if it is a path.ext then it
    will be converted to a full url (http://domain.com/relative/path.ext), and if
    it is a /full/path then it will be converted to http://domain.com/full/path

    based off of:
    https://github.com/waylan/Python-Markdown/blob/master/markdown/extensions/headerid.py
    """
    def __init__(self, post):
        self.config = {
            'post' : [post, 'the post instance this extension is working on'],
        }

    def extendMarkdown(self, md, md_globals):
        md.registerExtension(self)
        self.processor = HrefTreeprocessor()
        self.processor.md = md
        self.processor.config = self.getConfigs()
        #md.treeprocessors.add('href', self.processor, ">")
        md.treeprocessors['domevent'] = self.processor


class HrefTreeprocessor(Treeprocessor):
    """
    support for HrefExtension

    http://effbot.org/zone/pythondoc-elementtree-ElementTree.htm
    """
    URL_RE = re.compile(ur"^(?:https?:\/\/|\/\/)", re.I)

    def normalize_url(self, url):
        """normalizes the url into a full url"""
        m = self.URL_RE.search(url)
        if not m:
            post = self.config['post']
            config = post.config
            uri = post.uri
            base_url = config.base_url

            if url.startswith('/'):
                url = "{}{}".format(base_url, url)

            else:
                for f in post.directory.other_files:
                    basename = os.path.basename(f)
                    if url == basename:
                        url = "{}/{}".format(post.url, basename)

        return url

    def run(self, doc):
        for elem in doc.getiterator():
            if elem.tag == 'a':
                href = elem.get('href')
                url = self.normalize_url(href)
                elem.set('href', url)


class HrefExtension(Extension):
    """
    this will modify all a href attributes to be full url paths

    if it is an http://... url, nothing happens, if it is a path.ext then it
    will be converted to a full url (http://domain.com/relative/path.ext), and if
    it is a /full/path then it will be converted to http://domain.com/full/path

    based off of:
    https://github.com/waylan/Python-Markdown/blob/master/markdown/extensions/headerid.py
    """
    def __init__(self, post):
        self.config = {
            'post' : [post, 'the post instance this extension is working on'],
        }

    def extendMarkdown(self, md, md_globals):
        md.registerExtension(self)
        self.processor = HrefTreeprocessor()
        self.processor.md = md
        self.processor.config = self.getConfigs()
        #md.treeprocessors.add('href', self.processor, ">")
        md.treeprocessors['href'] = self.processor


class ImageTreeprocessor(HrefTreeprocessor):
    """support for ImageExtension"""
    def iterparent(self, tree):
        """
        http://effbot.org/zone/element.htm
        """
        for parent in tree.getiterator():
            for child in parent:
                yield parent, child

    def run(self, doc):
        post = self.config['post']
        config = post.config
        for parent_elem, elem in self.iterparent(doc):
            if elem.tag == 'img':
                class_attr = parent_elem.get('class')
                if not class_attr: class_attr = ''
                img_class = ''
                if len(parent_elem) == 1:
                    head_text = parent_elem.text
                    if not head_text: head_text = ''
                    head_text = head_text.strip()

                    tail_text = elem.tail
                    if not tail_text: tail_text = ''
                    tail_text = tail_text.strip()

                    if head_text or tail_text:
                        img_class += ' image-floating'

                    else:
                        img_class += ' image-centered'

                else:
                    img_class += ' image-floating'

                class_attr += img_class
                parent_elem.set('class', class_attr.strip())

                # also normalize the url
                src = elem.get('src')
                url = self.normalize_url(src)
                elem.set('src', url)


class ImageExtension(Extension):
    """
    this looks at img tags and makes sure they are centered if they are solo or floating
    if they are in paragraph's with content
    """
    def __init__(self, post):
        self.config = {
            'post' : [post, 'the post instance this extension is working on'],
        }

    def extendMarkdown(self, md, md_globals):
        md.registerExtension(self)
        self.processor = ImageTreeprocessor()
        self.processor.md = md
        self.processor.config = self.getConfigs()
        #md.treeprocessors.add('href', self.processor, ">")
        md.treeprocessors['image'] = self.processor


class HighlightExtension(fenced_code.FencedCodeExtension):
    """
    A compatible extenstion to highlight.js for python-markdown

    http://highlightjs.org
    https://github.com/isagalaev/highlight.js

    extends this markdown ext: http://pythonhosted.org/Markdown/extensions/fenced_code_blocks.html
    """
    def extendMarkdown(self, md, md_globals):
        """ Add FencedBlockPreprocessor to the Markdown instance. """
        md.registerExtension(self)

        md.preprocessors.add(
            'code_block',
            CodeBlockPreprocessor(md),
            ">normalize_whitespace"
        )


class CodeBlockFormatter(HtmlFormatter):
    """
    based off the example found here: http://pygments.org/docs/formatters/

    http://pygments.org/docs/
    """
    def wrap(self, source, outfile):
        return self._wrap_code(source)

    def _wrap_code(self, source):
        code_tag = '<code>'
        if self.cssclass:
            code_tag = '<code class="{}">'.format(self.cssclass)

        yield 0, '<pre>{}'.format(code_tag)
        for i, t in source:
            yield i, t
        yield 0, '</code></pre>'


class CodeBlockPreprocessor(fenced_code.FencedBlockPreprocessor):
    """
    Generates compatible code blocks that can be used with default highlight.js
    """
    FENCED_BLOCK_RE = re.compile(
        ur"^(?P<fence>(?:`{3,}))[ ]*(?P<lang>[a-z0-9_+-]*)[ ]*(?P<code>.*?)(?<=\n)(?P=fence)[ ]*$",
        re.MULTILINE | re.DOTALL | re.VERBOSE
    )

    def run(self, lines):
        """ Match and store Fenced Code Blocks in the HtmlStash. """
        text = u"\n".join(lines)
        while 1:
            m = self.FENCED_BLOCK_RE.search(text)
            if m:
                lang = 'no-highlight'
                if m.group('lang'):
                    lang = m.group('lang')

                try:
                    lexer = get_lexer_by_name(lang)

                except ValueError:
                    lexer = TextLexer()

                formatter = CodeBlockFormatter(cssclass="codeblock {}".format(lang))
                code = highlight(m.group('code'), lexer, formatter)

                placeholder = self.markdown.htmlStash.store(code, safe=True)
                text = u'{}\n{}\n{}'.format(text[:m.start()], placeholder, text[m.end():])

            else:
                break

        return text.split("\n")




import os
import re
from distutils import dir_util
import shutil
import types

from . import echo

class Directory(object):

    #content_file_regex = ur'\.(md|html|txt|markdown)$'
    content_file_regex = ur'\.(md|html|markdown)$'

    @property
    def basename(self):
        return os.path.basename(self.path)

    @property
    def content_file(self):
        for f in self.files(self.content_file_regex):
            break
        return f

    @property
    def other_files(self):
        for f in self.files():
            if not re.search(self.content_file_regex, f, re.I):
                yield f

    def __init__(self, *bits):
        self.path = ''
        self.ancestor_dir = None
        if bits:
            bits = list(bits)
            bits[0] = self.normalize(bits[0])
            for i in xrange(1, len(bits)):
                bits[i] = bits[i].strip('\\/')
            self.path = os.path.join(*bits)

    @classmethod
    def normalize(cls, d):
        """completely normalize a relative path (a path with ../, ./, or ~/)"""
        return os.path.abspath(os.path.expanduser(str(d)))

    def exists(self):
        return os.path.isdir(self.path)

    def copy_file(self, input_file):
        """copy the input_file to this directory"""
        basename = os.path.basename(input_file)
        output_file = os.path.join(self.path, basename)
        echo.out("copy file {} to {}", input_file, output_file)
        return shutil.copy(input_file, output_file)

    def create(self):
        """create the directory path"""
        echo.out("create dir: {}", self.path)
        return dir_util.mkpath(self.path)

    def clear(self):
        """this will clear a directory path of all files and folders"""
        # http://stackoverflow.com/a/1073382/5006
        echo.out("clearing {}", self.path)
        dir_util.mkpath(self.path)
        for root, dirs, files in os.walk(self.path, topdown=True):
            for td in dirs:
                shutil.rmtree(os.path.join(root, td))

            for tf in files:
                os.unlink(os.path.join(root, tf))

            break

        return True

    def __div__(self, bits):
        if isinstance(bits, types.StringTypes):
            bits = [bits]
        else:
            bits = list(bits)

        return Directory(self.path, *bits)

    def __truediv__(self, bits):
        return self.__div__(bits)

    def __str__(self):
        return self.path

    def __unicode__(self):
        return unicode(self.path)

    def __iter__(self):
        for root_dir, dirs, _ in os.walk(self.path, topdown=True):
            dirs.sort()
            for basename in dirs:
                d = Directory(root_dir, basename)
                d.ancestor_dir = self
                if not d.is_private():
                    yield d

    def files(self, regex=None):
        fs = []
        for root_dir, _, files in os.walk(self.path, topdown=True):
            for basename in files:
                if basename.startswith('_'): continue
                if regex and not re.search(regex, basename, re.I):
                    continue
                fs.append(os.path.join(root_dir, basename))
            break

        return fs

    def is_post(self):
        """return true if this directory has a blog post in it"""
        return not self.is_aux() and self.files(self.content_file_regex)

    def is_aux(self):
        """return True if this is a directory with an index.* file in it"""
        ret_bool = False
        if self.files(ur'^index{}'.format(self.content_file_regex)):
            ret_bool = True

        return ret_bool

    def is_private(self):
        basename = self.basename
        return basename.startswith('_')

    def relative(self, ancestor_dir=None):
        """
        returns the relative bits to the parent_dir

        example --
            d = Directory("/foo/bar/baz/che")
            d.relative("/foo/bar") # baz/che
            d.relative("/foo") # bar/baz/che

        ancestor_dir -- string|Directory -- the directory you want to return that self
            is a child of, if parent_dir is empty than it will use self.ancestor_dir
        return -- string -- the part of the path that is relative
        """
        if not ancestor_dir:
            ancestor_dir = self.ancestor_dir
        assert ancestor_dir, "no ancestor_dir found"

        relative = self.path.replace(str(ancestor_dir), '').strip(os.sep)
        return relative


class ProjectDirectory(Directory):
    def __init__(self, *bits):
        super(ProjectDirectory, self).__init__(*bits)

        self.template_dir = Directory(self.path, 'template')
        self.input_dir = Directory(self.path, 'input')

